注意
跳转至页面底部下载完整示例代码。
在 torch.compile 中使用用户定义的 Triton 内核#
创建日期:2024年4月19日 | 最后更新:2026年4月29日 | 最后验证:2024年11月5日
作者: Oguz Ulgen
用户定义的 Triton 内核可用于优化模型计算的特定部分。这些内核使用 Triton 语言编写,旨在更轻松地实现极致的硬件性能。通过在 torch.compile 中使用用户定义的 Triton 内核,你可以将这些优化后的计算整合到 PyTorch 模型中,从而可能获得显著的性能提升。
本教程演示了如何将用户定义的 Triton 内核与 torch.compile 结合使用。
先决条件#
在开始此秘籍之前,请确保您已具备以下条件
对
torch.compile和 Triton 有基本了解。请参考PyTorch 2.3 或更高版本
支持 Triton 的 GPU
import torch
from torch.utils._triton import has_triton
基本用法#
在本示例中,我们将结合 torch.compile 使用 Triton 文档中的简单向量加法内核。如需参考,请参阅 Triton 文档。
if not has_triton():
print("Skipping because triton is not supported on this device.")
else:
import triton
from triton import language as tl
@triton.jit
def add_kernel(
in_ptr0,
in_ptr1,
out_ptr,
n_elements,
BLOCK_SIZE: "tl.constexpr",
):
pid = tl.program_id(axis=0)
block_start = pid * BLOCK_SIZE
offsets = block_start + tl.arange(0, BLOCK_SIZE)
mask = offsets < n_elements
x = tl.load(in_ptr0 + offsets, mask=mask)
y = tl.load(in_ptr1 + offsets, mask=mask)
output = x + y
tl.store(out_ptr + offsets, output, mask=mask)
@torch.compile(fullgraph=True)
def add_fn(x, y):
output = torch.zeros_like(x)
n_elements = output.numel()
grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)
add_kernel[grid](x, y, output, n_elements, BLOCK_SIZE=4)
return output
x = torch.randn(4, device="cuda")
y = torch.randn(4, device="cuda")
out = add_fn(x, y)
print(f"Vector addition of\nX:\t{x}\nY:\t{y}\nis equal to\n{out}")
Vector addition of
X: tensor([ 0.5870, 0.4293, -0.1812, 0.4154], device='cuda:0')
Y: tensor([-1.3115, -0.6079, 0.7410, -0.2501], device='cuda:0')
is equal to
tensor([-0.7246, -0.1786, 0.5598, 0.1652], device='cuda:0')
高级用法#
Triton 的自动调优(autotune)功能是一个强大的工具,可以自动优化 Triton 内核的配置参数。它会探索一系列可能的配置,并选择最适合特定用例的配置以实现最佳性能。
当与 torch.compile 结合使用时,triton.autotune 可以帮助确保你的 PyTorch 模型尽可能高效地运行。以下是结合使用 torch.compile 和 triton.autotune 的示例。
注意
torch.compile 仅支持 triton.autotune 中的 configs、keys、restore_value 和 reset_to_zero 参数。
if not has_triton():
print("Skipping because triton is not supported on this device.")
else:
import triton
from triton import language as tl
@triton.autotune(
configs=[
triton.Config({"BLOCK_SIZE": 4}, num_stages=3, num_warps=8),
triton.Config({"BLOCK_SIZE": 4}, num_stages=4, num_warps=4),
triton.Config({"BLOCK_SIZE": 2}, num_stages=3, num_warps=8),
triton.Config({"BLOCK_SIZE": 2}, num_stages=4, num_warps=4),
],
key=[],
)
@triton.jit
def add_kernel_autotuned(
in_ptr0,
in_ptr1,
out_ptr,
n_elements,
BLOCK_SIZE: "tl.constexpr",
):
pid = tl.program_id(axis=0)
block_start = pid * BLOCK_SIZE
offsets = block_start + tl.arange(0, BLOCK_SIZE)
mask = offsets < n_elements
x = tl.load(in_ptr0 + offsets, mask=mask)
y = tl.load(in_ptr1 + offsets, mask=mask)
output = x + y
tl.store(out_ptr + offsets, output, mask=mask)
@torch.compile(fullgraph=True)
def add_fn(x, y):
output = torch.zeros_like(x)
n_elements = output.numel()
grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)
add_kernel_autotuned[grid](x, y, output, n_elements)
return output
x = torch.randn(4, device="cuda")
y = torch.randn(4, device="cuda")
out = add_fn(x, y)
print(f"Vector addition of\nX:\t{x}\nY:\t{y}\nis equal to\n{out}")
Vector addition of
X: tensor([-0.2110, 1.1786, -1.9163, 1.4693], device='cuda:0')
Y: tensor([ 1.1126, -0.7698, -1.7958, -0.2113], device='cuda:0')
is equal to
tensor([ 0.9016, 0.4089, -3.7121, 1.2581], device='cuda:0')
可组合性#
用户定义的 Triton 内核并不自动支持所有的 PyTorch 子系统。这体现在以下用例中:
添加 CPU 回退 (fallback)
添加
FlopCounter公式与张量子类 (Tensor Subclasses) 组合
要与更多的 PyTorch 子系统组合,请使用 torch.library.triton_op。
triton_op 是一种定义自定义算子的结构化方式,由一个或多个 Triton 内核提供支持:与常规自定义算子(torch.library.custom_op)类似,你可以通过 torch.library 指定与 PyTorch 子系统的交互。然而,与 torch.library.custom_op 不同(它对 torch.compile 而言是黑盒可调用对象),torch.compile 可以跟踪进入 triton_op 以应用优化。
以下是将 Triton 内核与 PyTorch 集成时所选 API 的对照表。
Triton 内核(无显式 |
|
|
|
|---|---|---|---|
支持推理 |
是 |
是 |
是 |
支持训练 |
在大多数情况下 |
是 |
是 |
支持 |
是 |
是 |
是 |
支持 |
在大多数情况下 |
在大多数情况下 |
在所有情况下 |
torch.compile 是否跟踪进入实现内部? |
是 |
是 |
否 |
支持 AOTInductor |
是 |
是 |
否 |
支持如 FlopCounterMode、CPU 回退、张量子类等 PyTorch 子系统 |
否 |
是 |
是 |
使用 triton_op 包装 Triton 内核#
使用 torch.library.triton_op 包装可能调用一个或多个 Triton 内核的函数。使用 torch.library.wrap_triton 包装对 Triton 内核的调用。
from torch.library import triton_op, wrap_triton
@triton_op("mylib::mysin", mutates_args={})
def mysin(x: torch.Tensor) -> torch.Tensor:
out = torch.empty_like(x)
n_elements = x.numel()
wrap_triton(sin_kernel)[(n_elements,)](x, out, n_elements, BLOCK_SIZE=4)
return out
@triton.jit
def sin_kernel(
in_ptr0,
out_ptr,
n_elements,
BLOCK_SIZE: "tl.constexpr",
):
pid = tl.program_id(axis=0)
block_start = pid * BLOCK_SIZE
offsets = block_start + tl.arange(0, BLOCK_SIZE)
mask = offsets < n_elements
x = tl.load(in_ptr0 + offsets, mask=mask)
output = tl.sin(x)
tl.store(out_ptr + offsets, output, mask=mask)
你可以通过以下两种方式之一调用 triton_op。
x = torch.randn(3, device="cuda")
y = mysin(x)
z = torch.ops.mylib.mysin.default(x)
assert torch.allclose(y, x.sin())
assert torch.allclose(z, x.sin())
由此产生的 triton_op 可与 torch.compile 和 AOTInductor 协同工作。
y = torch.compile(mysin)(x)
assert torch.allclose(y, x.sin())
添加训练支持#
使用 register_autograd 为 triton_op 添加自动微分公式。优先使用此方法,而不是 torch.autograd.Function(后者在与 torch.compile 结合时存在各种可组合性陷阱)。
请注意,反向传播必须是 PyTorch 可理解算子的组合。如果你希望反向传播也调用 Triton 内核,那么这些内核也必须包装在 triton_op 中。
@triton.jit
def cos_kernel(
in_ptr0,
out_ptr,
n_elements,
BLOCK_SIZE: "tl.constexpr",
):
pid = tl.program_id(axis=0)
block_start = pid * BLOCK_SIZE
offsets = block_start + tl.arange(0, BLOCK_SIZE)
mask = offsets < n_elements
x = tl.load(in_ptr0 + offsets, mask=mask)
output = tl.cos(x)
tl.store(out_ptr + offsets, output, mask=mask)
@triton_op("mylib::mycos", mutates_args={})
def mycos(x: torch.Tensor) -> torch.Tensor:
out = torch.empty_like(x)
n_elements = x.numel()
wrap_triton(cos_kernel)[(n_elements,)](x, out, n_elements, BLOCK_SIZE=4)
return out
def backward(ctx, grad):
x, = ctx.saved_tensors
return grad * mycos(x)
def setup_context(ctx, inputs, output):
x, = inputs
ctx.save_for_backward(x)
mysin.register_autograd(backward, setup_context=setup_context)
添加 CPU 回退#
Triton 内核无法在 CPU 上运行。使用 register_kernel 为 triton_op 添加 CPU(或任何其他设备)回退逻辑。
回退逻辑必须由 PyTorch 算子组成。
添加 FlopCounter 公式#
要指定 Triton 内核在 PyTorch 的浮点运算计数器中报告多少 FLOP,请使用 register_flop_formula。
from torch.utils.flop_counter import FlopCounterMode, register_flop_formula
@register_flop_formula(torch.ops.mylib.mysin)
def _(x_shape):
numel = 1
for s in x_shape:
numel *= s
return numel
x = torch.randn(3, device="cuda")
FlopCounterMode 需要 tabulate 库。在运行下面的代码之前,请确保已安装 tabulate,或运行 pip install tabulate 进行安装。
限制#
截至 PyTorch 2.3,torch.compile 对用户定义的 Triton 内核的支持包括动态形状、torch.autograd.Function、JIT Inductor 和 AOT Inductor。你可以将这些功能组合在一起构建复杂的高性能模型。
PyTorch 2.6 添加了 torch.library.triton_op,增加了对张量子类中用户定义 Triton 内核的支持以及其他高级特性。
然而,需要注意以下限制:
Triton 特性: 虽然
triton.heuristics可以独立使用,也可以在triton.autotune之前使用,但它不能在triton.autotune之后使用。这意味着如果需要同时使用triton.heuristics和triton.autotune,必须先使用triton.heuristics。
结论#
在本教程中,我们探讨了如何利用用户定义的 Triton 内核与 torch.compile。我们深入了解了简单向量加法内核的基本用法以及涉及 Triton 自动调优功能的高级用法。我们还讨论了用户定义 Triton 内核与其他 PyTorch 特性的可组合性,并指出了当前的一些限制。
另请参阅#
脚本运行总耗时: (0 分 3.302 秒)